/*	inline.h

	AFAPI PAPERS_JR inline primitive functions....

	This library is designed to be able to link two processors
	using either CAPERS (i.e., a "LapLink" cable) or PAPERS_JR.
	There are several differences between CAPERS and PAPERS_JR,
	but both are essentially passive connections in normal use.
	This, and being compatible with "LapLink" wiring, make the
	library different from that for TTL_PAPERS; however, PAPERS_JR
	is strictly a superset of CAPERS, so the same library can be
	used for both.  It will be a little hard to see the LEDs using
	CAPERS....  ;-)

	July 1996 by Hank Dietz
	August 1996 fixes by Tim Mattox & Hank Dietz
*/


#ifndef	_INLINE_H
#define	_INLINE_H


#include "afapi.h"

/*	p__inb() and p__outb()...

	These are the actual lowest-level hardware port I/O
	instructions used to touch the parallel port interface.  They
	are defined here, rather than borrowed from a library, to
	ensure that they don't add any delays...  many library
	versions insert code to allow port settle time, which we
	confirm using a handshake protocol instead.  Unfortunately,
	because PAPERS_JR and CAPERS do not contain any circuitry that
	can clean-up the PC port driving, and some ports do need a
	little help, we have added an option to do software debouncing
	on each port read.  This is enabled by defining P_DEBOUNCE in
	types.h.
*/


#ifdef	P_DEBOUNCE

extern inline
unsigned int
p___inb(unsigned short port)
{
	unsigned char _v;
__asm__ __volatile__ ("inb %w1,%b0"
		:"=a" (_v)
		:"d" (port), "0" (0));
	return _v;
}

extern inline
unsigned int
p__inb(unsigned short port)
{
	register unsigned int t = p___inb(port);

	while (t != p_inlast) {
		p_inlast = t;
		t = p___inb(port);
	}
	return(t);
}

#else

extern inline
unsigned int
p__inb(unsigned short port)
{
	unsigned char _v;
__asm__ __volatile__ ("inb %w1,%b0"
		:"=a" (_v)
		:"d" (port), "0" (0));
	return _v;
}

#endif


extern inline
void
p__outb(unsigned char value,
unsigned short port)
{
__asm__ __volatile__ ("outb %b0,%w1"
		:/* no outputs */
		:"a" (value), "d" (port));
}

/*	p__wait()...

	There is a bit of strangeness in PAPERS_JR in that there
	really isn't any aggregate function or barrier hardware.
	Don't ask how the following code "synchronizes" the two
	processors...  the answer is that it really doesn't, but
	behaves as though it did.  :-S  I said DO NOT ASK!

	Anyway, this code uses the same tricks used in the 960801
	TTL_PAPERS code to avoid conditionals and reduce code size.
*/

extern inline
void
p__wait(void)
{
	register unsigned char t;
	register unsigned char t1;
	register unsigned char match;

	/* Flip the single strobe bit... */
	t = ((t1 = p_outlast) ^ 0x10);

	/* PE0 goes first... */
	if (IPROC == 0) {
		p__outb(t, p_outport);
		t1 = t;
	}

	p_outlast = t;

	/* Wait for P_RDY state to match strobe state, flipping
	   the LED from green to RED if we need to poll three times
	*/
	match = ((t & 0x10) << 3);
	if ((p__inb(p_inport) & 0x80) == match) {
		if ((p__inb(p_inport) & 0x80) == match) {
			p__outb((t1 ^ (0x80 | 0x40)), p_outport);
			while ((p__inb(p_inport) & 0x80) == match) ;
			p__outb(t1, p_outport);
		}
	}

	/* PE1 responds... */
	if (IPROC != 0) {
		p__outb(t, p_outport);
	}
}

/*	p__swap()...

	With just two processors, all the world's a putGet...  and it
	is pretty obvious who we want to get from.  Call it p__swap().
*/	

extern inline
int
p__swap(register p_8u b)
{
	register unsigned char t;
	register unsigned char t1;
	register unsigned char match;
	register int result;

	/* Paste-in the new data and flip the strobe */
	t = ((((t1 = p_outlast) & 0xf0) | (b & 0x0f)) ^ 0x10);

	/* PE0 goes first... */
	if (IPROC == 0) {
		p__outb(t, p_outport);
		t1 = t;
	}

	p_outlast = t;

	/* Wait for P_RDY state to match strobe state, flipping
	   the LED from green to RED if we need to poll three times
	*/
	match = ((t & 0x10) << 3);
	if ((p__inb(p_inport) & 0x80) == match) {
		if ((p__inb(p_inport) & 0x80) == match) {
			p__outb((t1 ^ (0x80 | 0x40)), p_outport);
			while ((p__inb(p_inport) & 0x80) == match) ;
			p__outb(t1, p_outport);
		}
	}

	/* Read the data value again (time for data to settle) */
	result = p__inb(p_inport);

	/* PE1 responds... */
	if (IPROC != 0) {
		p__outb(t, p_outport);
	}

	/* The result is his value...  return with data bit vector
	   shifted into place and masked to 4 bits.
	*/
	return((result >> 3) & 0xf);
}

/*	p__nand()...

	Trivially built from p_swap()....
*/	

extern inline
int
p__nand(register p_8u b)
{
	return(0xf & ~(b & p__swap(b)));
}

/*	p__put() and p__get()...

	This code is very similar to p__swap(), but a few operations
	were not needed, so coding it separately allows them to be
	removed.  This helps a lot because many other library
	functions inline expand these two primitives.
*/	

extern inline
int
p__get(void)
{
	register unsigned char t;
	register unsigned char t1;
	register unsigned char match;
	register int result;

	/* Flip the strobe... */
	t = ((t1 = p_outlast) ^ 0x10);

	/* PE0 goes first... */
	if (IPROC == 0) {
		p__outb(t, p_outport);
		t1 = t;
	}

	p_outlast = t;

	/* Wait for P_RDY state to match strobe state, flipping
	   the LED from green to RED if we need to poll three times
	*/
	match = ((t & 0x10) << 3);
	if ((p__inb(p_inport) & 0x80) == match) {
		if ((p__inb(p_inport) & 0x80) == match) {
			p__outb((t1 ^ (0x80 | 0x40)), p_outport);
			while ((p__inb(p_inport) & 0x80) == match) ;
			p__outb(t1, p_outport);
		}
	}

	/* Read the data value again (time for data to settle) */
	result = p__inb(p_inport);

	/* PE1 responds... */
	if (IPROC != 0) {
		p__outb(t, p_outport);
	}

	/* The result is his value...  return with data bit vector
	   shifted into place and masked to 4 bits.
	*/
	return((result >> 3) & 0xf);
}


extern inline
void
p__put(register p_8u b)
{
	register unsigned char t;
	register unsigned char t1;
	register unsigned char match;
	register int result;

	/* Paste-in the new data and flip the strobe */
	t = ((((t1 = p_outlast) & 0xf0) | (b & 0x0f)) ^ 0x10);

	/* PE0 goes first... */
	if (IPROC == 0) {
		p__outb(t, p_outport);
		t1 = t;
	}

	p_outlast = t;

	/* Wait for P_RDY state to match strobe state, flipping
	   the LED from green to RED if we need to poll three times
	*/
	match = ((t & 0x10) << 3);
	if ((p__inb(p_inport) & 0x80) == match) {
		if ((p__inb(p_inport) & 0x80) == match) {
			p__outb((t1 ^ (0x80 | 0x40)), p_outport);
			while ((p__inb(p_inport) & 0x80) == match) ;
			p__outb(t1, p_outport);
		}
	}

	/* No need to read his data... */

	/* PE1 responds... */
	if (IPROC != 0) {
		p__outb(t, p_outport);
	}
}

/*	p__waitbar(), p__any(), and p_all()...

	These really are not bottom-level primitives for PAPERS_JR
	hardware, but each is small enough to inline....
*/

extern inline
p_bar
p__waitbar(register p_1u f)
{
	/* Not real hard, given that NPROC=2  ;-) */
	int pat[2];

	f = (f != 0);
	pat[IPROC] = f;
	pat[IPROC ^ 1] = p__swap(f);

	return((p_bar) (pat[0] | (pat[1] << 1)));
}


extern inline
p_1u
p__any(register p_1u f)
{
	return(p__swap(f != 0) || f);
}


extern inline
p_1u
p__all(register p_1u f)
{
	return(p__swap(f != 0) && f);
}

inline extern int
p__countbar(register p_bar b)
{
	return((b & 1) + ((b & 2) == 2));
}


inline extern int
p__firstbar(register p_bar b)
{
	if (b & 1) return(0);
	if (b & 2) return(1);
	return(2);
}

/*	Inline signal primitives...

	Signals are a bit funky here because the normal way that
	barriers are sent has PE0 examining PE1's value only after it
	has sent its own value...  this means that PE0 would not
	normally be able to detect that PE1 had initiated a signal.
	However, this can be taken care of simply enough by making
	sure that both PE0 and PE1 can look, and then letting PE0
	initiate the signal handling barrier....

	Anyway, what that all boils down to is that both PEs set all
	their data bits to 0 and we just signal on the data bits.
	However, unlike TTL_PAPERS, you don't see your own signal.
*/

extern inline void
s__signal(register int s)
{
	/* This is not worth inlining... */
	s_signal(s);
}


extern inline void
s__poll(void)
{
	/* Basically need to se if he wants us...  because we don't
	   have a signal to send if we are here.
	*/
	if (p__inb(p_inport) & 0x38) {
		s_signal(S_SIGANY);
	}
}


extern inline void
s__wait(void)
{
	/* Just exit and re-enter... that flushes things */
	s_exit();
	s_init();
}

/*	Inline RSM interface...

	It happens that only the signal handler itself is a lot of
	code, thus, these can also be inlined....
*/

extern inline void
s__update(register void *a,
register int n)
{
	s_updateArgs.a = a;
	s_updateArgs.n = n;
	s__signal(S_SIGMEM);
}


extern inline void
s__initLock(register s_lock *lock)
{
	/* Initialize lock... */
	*lock = S_INITLOCK;
	s__update(((void *) lock), sizeof(s_lock));
}


extern inline void
s__acquire(register s_lock *lock)
{
	/* Acquire lock...  the funky thing about this is that it
	   works for our system BECAUSE replicated shared memory
	   updates are internally executed in a purely synchronous
	   way...  it wouldn't work using the SMP hardware directly.
	*/
	register s_lock us = (IPROC + 1);

	do {
		/* Wait for lock to be open... */
		while (*lock != S_INITLOCK) {
			s__poll();
		}

		/* Try to grab it for us */
		*lock = us;
		s__update(((void *) lock), sizeof(s_lock));

		/* Continue looping if we lost the race... */
	} while (*lock != us);
}


extern inline void
s__release(register s_lock *lock)
{
	/* Unlock lock.
	   Hmm...  this looks like s__initLock(), doesn't it?
	*/
	*lock = S_INITLOCK;
	s__update(((void *) lock), sizeof(s_lock));
}


#endif
